class Reverb:
    pass
class ScatterLocation:
    pass
class Chorus:
    pass
class Vocoder:
    pass
class ShuffleSound:
    pass
class LPFilter:
    pass
class Notch:
    pass
class Notch:
    pass
class Wavelet:
    pass
class Delays:
    pass
